#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The agri_l1 reader tests package."""

import os
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from satpy.tests.reader_tests.test_hdf5_utils import FakeHDF5FileHandler

ALL_BAND_NAMES = ["C01", "C02", "C03", "C04", "C05", "C06", "C07",
                  "C08", "C09", "C10", "C11", "C12", "C13", "C14"]

CHANNELS_BY_RESOLUTION = {500: ["C02"],
                          1000: ["C01", "C02", "C03"],
                          2000: ["C01", "C02", "C03", "C04", "C05", "C06", "C07"],
                          4000: ALL_BAND_NAMES,
                          "GEO": "solar_azimuth_angle"
                          }

RESOLUTION_LIST = [500, 1000, 2000, 4000]

AREA_EXTENTS_BY_RESOLUTION = {"FY4A": {
    500: (-5496021.008869, 5495021.005046, -5493520.999312, 5496021.008869),
    1000: (-5496021.076004, 5494021.068334, -5491021.05683, 5496021.076004),
    2000: (-5496021.210274, 5492021.194837, -5486021.171682, 5496021.210274),
    4000: (-5496021.210274, 5488021.1794, -5476021.13309, 5496021.210274)
},
    "FY4B": {
        500: (-5496021.008869, 5495021.005046, -5493520.999312, 5496021.008869),
        1000: (-5496021.076004, 5494021.068334, -5491021.05683, 5496021.076004),
        2000: (-5496021.210274, 5492021.194837, -5486021.171682, 5496021.210274),
        4000: (-5496021.210274, 5488021.1794, -5476021.13309, 5496021.210274)
    }}


class FakeHDF5FileHandler2(FakeHDF5FileHandler):
    """Swap-in HDF5 File Handler."""

    def _make_cal_data(self, cwl, ch, dims):
        """Make test data."""
        return xr.DataArray(
            da.from_array((np.arange(10.) + 1.) / 10., [dims[0] * dims[1]]),
            attrs={
                "Slope": np.array(1.), "Intercept": np.array(0.),
                "FillValue": np.array(-65535.0),
                "units": "NUL",
                "center_wavelength": "{}um".format(cwl).encode("utf-8"),
                "band_names": "band{}(band number is range from 1 to 14)"
                .format(ch).encode("utf-8"),
                "long_name": "Calibration table of {}um Channel".format(cwl).encode("utf-8"),
                "valid_range": np.array([0, 1.5]),
            },
            dims="_const")

    def _make_nom_data(self, cwl, ch, dims):
        # Add +1 to check that values beyond the LUT are clipped
        data_np = np.arange(10, dtype=np.uint16).reshape((2, 5)) + 1
        fill_value = 65535
        valid_max = 4095
        if ch == 7:
            # mimic C07 bug where the fill value is in the LUT
            fill_value = 9  # at index [1, 3] (second to last element)
            valid_max = 8
        return xr.DataArray(
            da.from_array(data_np, chunks=[dim for dim in dims]),
            attrs={
                "Slope": np.array(1.), "Intercept": np.array(0.),
                "FillValue": np.array(fill_value),
                "units": "DN",
                "center_wavelength": "{}um".format(cwl).encode("utf-8"),
                "band_names": "band{}(band number is range from 1 to 14)"
                .format(ch).encode("utf-8"),
                "long_name": "Calibration table of {}um Channel".format(cwl).encode("utf-8"),
                "valid_range": np.array([0, valid_max]),
            },
            dims=("_RegLength", "_RegWidth"))

    def _make_geo_data(self, dims):
        return xr.DataArray(
            da.from_array(np.arange(0., 360., 36., dtype=np.float32).reshape((2, 5)),
                          [dim for dim in dims]),
            attrs={
                "Slope": np.array(1.), "Intercept": np.array(0.),
                "FillValue": np.array(65535.),
                "units": "NUL",
                "band_names": "NUL",
                "valid_range": np.array([0., 360.]),
            },
            dims=("_RegLength", "_RegWidth"))

    def _create_coeffs_array(self, channel_numbers: list[int]) -> xr.DataArray:
        # make coefficients consistent between file types
        all_possible_coeffs = (np.arange(14 * 2).reshape((14, 2)) + 1.0) / np.array([1E4, 1E2])
        # get the coefficients for the specific channels this resolution has
        these_coeffs = all_possible_coeffs[[chan_num - 1 for chan_num in channel_numbers]]
        data = xr.DataArray(
            da.from_array(these_coeffs, chunks=[len(channel_numbers), 2]),
            attrs={
                "Slope": 1., "Intercept": 0.,
                "FillValue": 0,
                "units": "NUL",
                "band_names": "NUL",
                "long_name": b"Calibration coefficient (SCALE and OFFSET)",
                "valid_range": [-500, 500],
            },
            dims=("_num_channel", "_coefs"))
        return data

    def _create_channel_data(self, chs, cwls):
        dim_0 = 2
        dim_1 = 5
        data = {}
        for chan_num, chan_wl in zip(chs, cwls):
            cal_data = self._make_cal_data(chan_wl, chan_num, [dim_0, dim_1])
            data[f"CALChannel{chan_num:02d}"] = cal_data
            data[f"Calibration/CALChannel{chan_num:02d}"] = cal_data
            nom_data = self._make_nom_data(chan_wl, chan_num, [dim_0, dim_1])
            data[f"NOMChannel{chan_num:02d}"] = nom_data
            data[f"Data/NOMChannel{chan_num:02d}"] = nom_data
        data["CALIBRATION_COEF(SCALE+OFFSET)"] = self._create_coeffs_array(chs)
        data["Calibration/CALIBRATION_COEF(SCALE+OFFSET)"] = self._create_coeffs_array(chs)
        return data

    def _get_500m_data(self):
        chs = [2]
        cwls = [0.65]
        return self._create_channel_data(chs, cwls)

    def _get_1km_data(self):
        chs = [1, 2, 3]
        cwls = [0.47, 0.65, 0.83]
        return self._create_channel_data(chs, cwls)

    def _get_2km_data(self):
        chs = [1, 2, 3, 4, 5, 6, 7]
        cwls = [0.47, 0.65, 0.83, 1.37, 1.61, 2.22, 3.72]
        return self._create_channel_data(chs, cwls)

    def _get_4km_data(self):
        chs = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
        cwls = [0.47, 0.65, 0.83, 1.37, 1.61, 2.22, 3.72, 3.72, 6.25, 7.10, 8.50, 10.8, 12, 13.5]
        return self._create_channel_data(chs, cwls)

    def _get_geo_data(self):
        dim_0 = 2
        dim_1 = 5
        data = {"NOMSunAzimuth": self._make_geo_data([dim_0, dim_1]),
                "Navigation/NOMSunAzimuth": self._make_geo_data([dim_0, dim_1])}
        return data

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        global_attrs = {
            "/attr/NOMCenterLat": np.array(0.0),
            "/attr/NOMCenterLon": np.array(104.7),
            "/attr/NOMSatHeight": np.array(42164140.0),
            "/attr/dEA": np.array(6378.14),
            "/attr/dObRecFlat": np.array(298.257223563),
            "/attr/OBIType": "REGC",
            "/attr/RegLength": np.array(2.0),
            "/attr/RegWidth": np.array(5.0),
            "/attr/Begin Line Number": np.array(0),
            "/attr/End Line Number": np.array(1),
            "/attr/Begin Pixel Number": np.array(0),
            "/attr/End Pixel Number": np.array(1),
            "/attr/Observing Beginning Date": "2019-06-03", "/attr/Observing Beginning Time": "00:30:01.807",
            "/attr/Observing Ending Date": "2019-06-03", "/attr/Observing Ending Time": "00:34:07.572",
            "/attr/Satellite Name": "FY4A", "/attr/Sensor Identification Code": "AGRI", "/attr/Sensor Name": "AGRI",
        }

        data = {}
        if self.filetype_info["file_type"] == "agri_l1_0500m":
            data = self._get_500m_data()
        elif self.filetype_info["file_type"] == "agri_l1_1000m":
            data = self._get_1km_data()
        elif self.filetype_info["file_type"] == "agri_l1_2000m":
            data = self._get_2km_data()
            global_attrs["/attr/Observing Beginning Time"] = "00:30:01"
            global_attrs["/attr/Observing Ending Time"] = "00:34:07"
        elif self.filetype_info["file_type"] == "agri_l1_4000m":
            data = self._get_4km_data()
        elif self.filetype_info["file_type"] == "agri_l1_4000m_geo":
            data = self._get_geo_data()

        test_content = {}
        test_content.update(global_attrs)
        test_content.update(data)

        return test_content


def _create_filenames_from_resolutions(satname, *resolutions):
    """Create filenames from the given resolutions."""
    if "GEO" in resolutions:
        return [f"{satname}-_AGRI--_N_REGC_1047E_L1-_GEO-_MULT_NOM_20190603003000_20190603003416_4000M_V0001.HDF"]
    pattern = (f"{satname}-_AGRI--_N_REGC_1047E_L1-_FDI-_MULT_NOM_20190603003000_20190603003416_"
               "{resolution:04d}M_V0001.HDF")
    return [pattern.format(resolution=resolution) for resolution in resolutions]


class Test_HDF_AGRI_L1_cal:
    """Test VIRR L1B Reader."""

    yaml_file = "agri_fy4a_l1.yaml"

    def setup_method(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.agri_l1 import HDF_AGRI_L1
        from satpy.readers.core.fy4 import FY4Base
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.fy4 = mock.patch.object(FY4Base, "__bases__", (FakeHDF5FileHandler2,))
        self.p = mock.patch.object(HDF_AGRI_L1.__class__, (self.fy4,))
        self.fake_handler = self.fy4.start()
        self.p.is_local = True
        self.satname = "FY4A"

        self.expected = {
            1: np.array([[2.01, 2.02, 2.03, 2.04, 2.05], [2.06, 2.07, 2.08, 2.09, 2.1]]),
            2: np.array([[4.03, 4.06, 4.09, 4.12, 4.15], [4.18, 4.21, 4.24, 4.27, 4.3]]),
            3: np.array([[6.05, 6.1, 6.15, 6.2, 6.25], [6.3, 6.35, 6.4, 6.45, 6.5]]),
            4: np.array([[8.07, 8.14, 8.21, 8.28, 8.35], [8.42, 8.49, 8.56, 8.63, 8.7]]),
            5: np.array([[10.09, 10.18, 10.27, 10.36, 10.45], [10.54, 10.63, 10.72, 10.81, 10.9]]),
            6: np.array([[12.11, 12.22, 12.33, 12.44, 12.55], [12.66, 12.77, 12.88, 12.99, 13.1]]),
            7: np.array([[0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, np.nan, np.nan]]),
            8: np.array([[0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1., np.nan]]),
            9: np.array([[0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1., np.nan]]),
            10: np.array([[0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1., np.nan]]),
            11: np.array([[0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1., np.nan]]),
            12: np.array([[0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1., np.nan]]),
            13: np.array([[0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1., np.nan]]),
            14: np.array([[0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1., np.nan]])
        }

    def teardown_method(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()

    def test_times_correct(self):
        """Test that the reader handles the two possible time formats correctly."""
        reader = self._create_reader_for_resolutions(1000)
        np.testing.assert_almost_equal(reader.start_time.microsecond, 807000)
        reader = self._create_reader_for_resolutions(2000)
        np.testing.assert_almost_equal(reader.start_time.microsecond, 0)

    def test_fy4a_channels_are_loaded_with_right_resolution(self):
        """Test all channels are loaded with the right resolution."""
        reader = self._create_reader_for_resolutions(*RESOLUTION_LIST)

        available_datasets = reader.available_dataset_ids

        for resolution_to_test in RESOLUTION_LIST:
            self._check_keys_for_dsq(available_datasets, resolution_to_test)

    def test_agri_all_bands_have_right_units(self):
        """Test all bands have the right units."""
        reader = self._create_reader_for_resolutions(*RESOLUTION_LIST)

        band_names = ALL_BAND_NAMES
        res = reader.load(band_names)
        assert len(res) == 14

        for band_name in band_names:
            assert res[band_name].shape == (2, 5)
            self._check_units(band_name, res)

    def test_agri_orbital_parameters_are_correct(self):
        """Test orbital parameters are set correctly."""
        reader = self._create_reader_for_resolutions(*RESOLUTION_LIST)

        band_names = ALL_BAND_NAMES
        res = reader.load(band_names)

        # check whether the data type of orbital_parameters is float
        orbital_parameters = res[band_names[0]].attrs["orbital_parameters"]
        for attr in orbital_parameters:
            assert isinstance(orbital_parameters[attr], float)
        assert orbital_parameters["satellite_nominal_latitude"] == 0.
        assert orbital_parameters["satellite_nominal_longitude"] == 104.7
        assert orbital_parameters["satellite_nominal_altitude"] == 42164140.0

    @staticmethod
    def _check_keys_for_dsq(available_datasets, resolution_to_test):
        from satpy.dataset.data_dict import get_key
        from satpy.tests.utils import make_dsq

        band_names = CHANNELS_BY_RESOLUTION[resolution_to_test]
        for band_name in band_names:
            ds_q = make_dsq(name=band_name, resolution=resolution_to_test)
            res = get_key(ds_q, available_datasets, num_results=0, best=False)
            if band_name < "C07":
                assert len(res) == 2
            else:
                assert len(res) == 3

    def test_agri_counts_calibration(self):
        """Test loading data at counts calibration."""
        from satpy.tests.utils import make_dsq
        reader = self._create_reader_for_resolutions(*RESOLUTION_LIST)

        ds_ids = []
        band_names = CHANNELS_BY_RESOLUTION[4000]
        for band_name in band_names:
            ds_ids.append(make_dsq(name=band_name, calibration="counts"))
        res = reader.load(ds_ids)
        assert len(res) == 14

        for band_name in band_names:
            assert res[band_name].shape == (2, 5)
            assert res[band_name].attrs["calibration"] == "counts"
            assert res[band_name].dtype == np.uint16
            assert res[band_name].attrs["units"] == "1"

    @pytest.mark.parametrize("satname", ["FY4A", "FY4B"])
    def test_agri_geo(self, satname):
        """Test loading data for angles."""
        from satpy.tests.utils import make_dsq
        self.satname = satname
        reader = self._create_reader_for_resolutions("GEO")
        band_name = "solar_azimuth_angle"
        ds_ids = [make_dsq(name=band_name)]
        res = reader.load(ds_ids)
        assert len(res) == 1

        np.testing.assert_almost_equal(np.nanmin(res[band_name]), 0.)
        np.testing.assert_almost_equal(np.nanmax(res[band_name]), 324.)

        assert res[band_name].shape == (2, 5)
        assert res[band_name].dtype == np.float32

    def _create_reader_for_resolutions(self, *resolutions):
        from satpy.readers.core.loading import load_reader
        filenames = _create_filenames_from_resolutions(self.satname, *resolutions)
        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        assert len(filenames) == len(files)
        reader.create_filehandlers(files)
        # Make sure we have some files
        assert reader.file_handlers
        return reader

    @pytest.mark.parametrize("resolution_to_test", RESOLUTION_LIST)
    @pytest.mark.parametrize("satname", ["FY4A", "FY4B"])
    def test_agri_for_one_resolution(self, resolution_to_test, satname):
        """Test loading data when only one resolution is available."""
        self.satname = satname
        reader = self._create_reader_for_resolutions(resolution_to_test)

        available_datasets = reader.available_dataset_ids
        band_names = CHANNELS_BY_RESOLUTION[resolution_to_test]
        self._assert_which_channels_are_loaded(available_datasets, band_names, resolution_to_test)
        res = reader.load(band_names)
        assert len(res) == len(band_names)
        self._check_calibration_and_units(band_names, res)
        for band_name in band_names:
            np.testing.assert_allclose(res[band_name].attrs["area"].area_extent,
                                       AREA_EXTENTS_BY_RESOLUTION[satname][resolution_to_test])

    def _check_calibration_and_units(self, band_names, result):
        for band_name in band_names:
            band_number = int(band_name[-2:])
            assert result[band_name].attrs["sensor"].islower()
            assert result[band_name].shape == (2, 5)
            np.testing.assert_allclose(result[band_name].values, self.expected[band_number], equal_nan=True)
            self._check_units(band_name, result)

    @staticmethod
    def _check_units(band_name, result):
        if band_name < "C07":
            assert result[band_name].attrs["calibration"] == "reflectance"
        else:
            assert result[band_name].attrs["calibration"] == "brightness_temperature"
        if band_name < "C07":
            assert result[band_name].attrs["units"] == "%"
        else:
            assert result[band_name].attrs["units"] == "K"

    @staticmethod
    def _assert_which_channels_are_loaded(available_datasets, band_names, resolution_to_test):
        from satpy.dataset.data_dict import get_key
        from satpy.tests.utils import make_dsq

        other_resolutions = RESOLUTION_LIST.copy()
        other_resolutions.remove(resolution_to_test)
        for band_name in band_names:
            for resolution in other_resolutions:
                ds_q = make_dsq(name=band_name, resolution=resolution)
                with pytest.raises(KeyError):
                    _ = get_key(ds_q, available_datasets, num_results=0, best=False)

            ds_q = make_dsq(name=band_name, resolution=resolution_to_test)
            res = get_key(ds_q, available_datasets, num_results=0, best=False)
            if band_name < "C07":
                assert len(res) == 2
            else:
                assert len(res) == 3



# Copyright (c) 2016-2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.meris_nc_sen3 module."""
import unittest
import unittest.mock as mock


class TestMERISReader(unittest.TestCase):
    """Test various meris_nc_sen3 filehandlers."""

    @mock.patch("xarray.open_dataset")
    def test_instantiate(self, mocked_dataset):
        """Test initialization of file handlers."""
        from satpy.readers.meris_nc_sen3 import NCMERIS2, NCMERISCal, NCMERISGeo
        from satpy.tests.utils import make_dataid

        ds_id = make_dataid(name="M01", calibration="reflectance")
        ds_id2 = make_dataid(name="wsqf", calibration="reflectance")
        filename_info = {"mission_id": "ENV", "dataset_name": "M01", "start_time": 0, "end_time": 0}

        test = NCMERISCal("somedir/somefile.nc", filename_info, "c")
        test.get_dataset(ds_id, filename_info)
        mocked_dataset.assert_called()
        mocked_dataset.reset_mock()

        test = NCMERISGeo("somedir/somefile.nc", filename_info, "c")
        test.get_dataset(ds_id, filename_info)
        mocked_dataset.assert_called()
        mocked_dataset.reset_mock()

        test = NCMERIS2("somedir/somefile.nc", filename_info, "c")
        test.get_dataset(ds_id, {"nc_key": "the_key"})
        test.get_dataset(ds_id2, {"nc_key": "the_key"})
        mocked_dataset.assert_called()
        mocked_dataset.reset_mock()

    @mock.patch("xarray.open_dataset")
    def test_open_file_objects(self, mocked_open_dataset):
        """Test initialization of file handlers."""
        from satpy.readers.olci_nc import NCOLCIBase
        filename_info = {"mission_id": "ENV", "dataset_name": "M01", "start_time": 0, "end_time": 0}

        open_file = mock.MagicMock()

        file_handler = NCOLCIBase(open_file, filename_info, "c")
        #  deepcode ignore W0104: This is a property that is actually a function call.
        file_handler.nc  # pylint: disable=W0104
        mocked_open_dataset.assert_called()
        open_file.open.assert_called()
        assert (open_file.open.return_value in mocked_open_dataset.call_args[0] or
                open_file.open.return_value == mocked_open_dataset.call_args[1].get("filename_or_obj"))

    @mock.patch("xarray.open_dataset")
    def test_get_dataset(self, mocked_dataset):
        """Test reading datasets."""
        import numpy as np
        import xarray as xr

        from satpy.readers.meris_nc_sen3 import NCMERIS2
        from satpy.tests.utils import make_dataid
        mocked_dataset.return_value = xr.Dataset({"mask": (["rows", "columns"],
                                                           np.array([1 << x for x in range(30)]).reshape(5, 6))},
                                                 coords={"rows": np.arange(5),
                                                         "columns": np.arange(6)})
        ds_id = make_dataid(name="mask")
        filename_info = {"mission_id": "ENV", "dataset_name": "mask", "start_time": 0, "end_time": 0}
        test = NCMERIS2("somedir/somefile.nc", filename_info, "c")
        res = test.get_dataset(ds_id, {"nc_key": "mask"})
        assert res.dtype == np.dtype("bool")

    @mock.patch("xarray.open_dataset")
    def test_meris_angles(self, mocked_dataset):
        """Test reading datasets."""
        import numpy as np
        import xarray as xr

        from satpy.readers.meris_nc_sen3 import NCMERISAngles
        from satpy.tests.utils import make_dataid
        attr_dict = {
            "ac_subsampling_factor": 1,
            "al_subsampling_factor": 2,
        }
        mocked_dataset.return_value = xr.Dataset({"SAA": (["tie_rows", "tie_columns"],
                                                          np.array([1 << x for x in range(30)]).reshape(5, 6)),
                                                  "SZA": (["tie_rows", "tie_columns"],
                                                          np.array([1 << x for x in range(30)]).reshape(5, 6)),
                                                  "OAA": (["tie_rows", "tie_columns"],
                                                          np.array([1 << x for x in range(30)]).reshape(5, 6)),
                                                  "OZA": (["tie_rows", "tie_columns"],
                                                          np.array([1 << x for x in range(30)]).reshape(5, 6))},
                                                 coords={"rows": np.arange(5),
                                                         "columns": np.arange(6)},
                                                 attrs=attr_dict)
        filename_info = {"mission_id": "ENV", "dataset_name": "M01", "start_time": 0, "end_time": 0}

        ds_id = make_dataid(name="solar_azimuth_angle")
        ds_id2 = make_dataid(name="satellite_zenith_angle")
        test = NCMERISAngles("somedir/somefile.nc", filename_info, "c")
        test.get_dataset(ds_id, filename_info)
        test.get_dataset(ds_id2, filename_info)
        mocked_dataset.assert_called()
        mocked_dataset.reset_mock()

    @mock.patch("xarray.open_dataset")
    def test_meris_meteo(self, mocked_dataset):
        """Test reading datasets."""
        import numpy as np
        import xarray as xr

        from satpy.readers.meris_nc_sen3 import NCMERISMeteo
        from satpy.tests.utils import make_dataid
        attr_dict = {
            "ac_subsampling_factor": 1,
            "al_subsampling_factor": 2,
        }
        data = {"humidity": (["tie_rows", "tie_columns"],
                             np.array([1 << x for x in range(30)]).reshape(5, 6)),
                "total_ozone": (["tie_rows", "tie_columns"],
                                np.array([1 << x for x in range(30)]).reshape(5, 6)),
                "sea_level_pressure": (["tie_rows", "tie_columns"],
                                       np.array([1 << x for x in range(30)]).reshape(5, 6)),
                "total_columnar_water_vapour": (["tie_rows", "tie_columns"],
                                                np.array([1 << x for x in range(30)]).reshape(5, 6))}
        mocked_dataset.return_value = xr.Dataset(data,
                                                 coords={"rows": np.arange(5),
                                                         "columns": np.arange(6)},
                                                 attrs=attr_dict)
        filename_info = {"mission_id": "ENV", "dataset_name": "humidity", "start_time": 0, "end_time": 0}

        ds_id = make_dataid(name="humidity")
        ds_id2 = make_dataid(name="total_ozone")
        test = NCMERISMeteo("somedir/somefile.nc", filename_info, "c")
        test.get_dataset(ds_id, filename_info)
        test.get_dataset(ds_id2, filename_info)
        mocked_dataset.assert_called()
        mocked_dataset.reset_mock()


class TestBitFlags(unittest.TestCase):
    """Test the bitflag reading."""

    def test_bitflags(self):
        """Test the BitFlags class."""
        from functools import reduce

        import numpy as np

        from satpy.readers.olci_nc import BitFlags

        flag_list = ["SEA_ICE", "MEGLINT", "HIGHGLINT", "CASE2_S", "CASE2_ANOM",
                     "HAZE_OVER_WATER", "WHITECAPS", "AC_FAIL", "BPAC_ON", "WHITE_SCATT",
                     "LOWRW", "HIGHRW", "OUT_OF_RANGE_AAC", "OUT_OF_SCOPE_AAC",
                     "OUT_OF_RANGE_OC_NN", "OUT_OF_SCOPE_OC_NN",
                     "OUT_OF_RANGE_CHL_OC4ME_INPUT", "OUT_OF_RANGE_CHL_OC4ME"]

        bits = np.array([1 << x for x in range(len(flag_list))])

        bflags = BitFlags(bits, flag_list=flag_list)

        items = ["SEA_ICE", "MEGLINT", "HIGHGLINT",
                 "HAZE_OVER_WATER", "WHITECAPS", "AC_FAIL", "WHITE_SCATT",
                 "LOWRW", "HIGHRW", "OUT_OF_RANGE_AAC", "OUT_OF_SCOPE_AAC",
                 "OUT_OF_RANGE_OC_NN", "OUT_OF_SCOPE_OC_NN",
                 "OUT_OF_RANGE_CHL_OC4ME_INPUT", "OUT_OF_RANGE_CHL_OC4ME"]

        mask = reduce(np.logical_or, [bflags[item] for item in items])
        expected = np.array([True, True, True, False, False, True, True,
                             True, False, True, True, True, True, True, True,
                             True, True, True])
        assert all(mask == expected)



# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""IASI L2 files."""

import datetime as dt
import logging

import dask.array as da
import h5py
import numpy as np
import xarray as xr

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.netcdf import NetCDF4FsspecFileHandler
from satpy.utils import get_legacy_chunk_size

CHUNK_SIZE = get_legacy_chunk_size()


# Scan timing values taken from
# http://oiswww.eumetsat.org/WEBOPS/eps-pg/IASI-L1/IASIL1-PG-4ProdOverview.htm
# Time between each scan in one scanline [ms]
SCAN_STEP_TIME = 8. / 37.
# Duration of one measurement [ms]
SCAN_STARE_DURATION = 151.0
# Time correction used between each 4-footprint measurements
VIEW_TIME_ADJUSTMENT = SCAN_STEP_TIME + SCAN_STARE_DURATION / 2.

VALUES_PER_SCAN_LINE = 120

# Epoch for the dates
EPOCH = dt.datetime(2000, 1, 1)

SHORT_NAMES = {"M01": "Metop-B",
               "M02": "Metop-A",
               "M03": "Metop-C"}

# Data in the "PWLR" group
DSET_NAMES = {
    "ozone_mixing_ratio": "O",
    "ozone_mixing_ratio_quality": "QO",
    "pressure": "P",
    "pressure_quality": "QP",
    "temperature": "T",
    "temperature_quality": "QT",
    "water_mixing_ratio": "W",
    "water_mixing_ratio_quality": "QW",
    "water_total_column": "WC",
    "ozone_total_column": "OC",
    "surface_skin_temperature": "Ts",
    "surface_skin_temperature_quality": "QTs",
    "emissivity": "E",
    "emissivity_quality": "QE",
}
# Data in "INFO" group
INFO_DSET_NAMES = {
    "amsu_instrument_flags": "FLG_AMSUBAD",
    "iasi_instrument_flags": "FLG_IASIBAD",
    "mhs_instrument_flags": "FLG_MHSBAD",
    "observation_minus_calculation": "OmC",
}
# Data in "Maps" group
MAPS_DSET_NAMES = {
    "surface_elevation": "Height",
    "surface_elevation_std": "HeightStd",
}

ALL_DATASET_NAMES = tuple(DSET_NAMES.keys()) + tuple(INFO_DSET_NAMES.keys()) + tuple(MAPS_DSET_NAMES.keys())

# Data in "L1C" group
GEO_NAMES = {"latitude": "Latitude",
             "longitude": "Longitude",
             "satellite_azimuth_angle": "SatAzimuth",
             "satellite_zenith_angle": "SatZenith",
             "sensing_time": {"day": "SensingTime_day",
                              "msec": "SensingTime_msec"},
             "solar_azimuth_angle": "SunAzimuth",
             "solar_zenith_angle": "SunZenith"}


LOGGER = logging.getLogger(__name__)


class IASIL2HDF5(BaseFileHandler):
    """File handler for IASI L2 HDF5 files."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init the file handler."""
        super(IASIL2HDF5, self).__init__(filename, filename_info,
                                         filetype_info)

        self.finfo = filename_info
        self.lons = None
        self.lats = None
        self.sensor = "iasi"
        short_name = filename_info["platform_id"]
        self.platform_name = SHORT_NAMES.get(short_name, short_name)

        self.mda = {}
        self.mda["platform_name"] = self.platform_name
        self.mda["sensor"] = "iasi"

    @property
    def start_time(self):
        """Get the start time."""
        return self.finfo["start_time"]

    @property
    def end_time(self):
        """Get the end time."""
        end_time = dt.datetime.combine(self.start_time.date(),
                                       self.finfo["end_time"].time())
        if end_time < self.start_time:
            end_time += dt.timedelta(days=1)
        return end_time

    def get_dataset(self, key, info):
        """Load a dataset."""
        with h5py.File(self.filename, "r") as fid:
            LOGGER.debug("Reading %s.", key["name"])
            if key["name"] in ALL_DATASET_NAMES:
                m_data = read_dataset(fid, key)
            else:
                m_data = read_geo(fid, key)
        m_data.attrs.update(info)
        m_data.attrs["sensor"] = self.sensor
        m_data.attrs["platform_name"] = self.platform_name

        return m_data


def read_dataset(fid, key):
    """Read dataset."""
    names, group = _get_names_and_group(key)
    dsid = names[key["name"]]
    dset = fid[group + dsid]
    if dset.ndim == 3:
        dims = ["y", "x", "level"]
    else:
        dims = ["y", "x"]
    data = xr.DataArray(da.from_array(_harmonize_data(dset[()]), chunks=CHUNK_SIZE),
                        name=key["name"], dims=dims)
    if data.dtype == np.float32:
        data = xr.where(data > 1e30, np.nan, data)

    dset_attrs = dict(dset.attrs)
    data.attrs.update(dset_attrs)

    return data


def _get_names_and_group(key):
    name = key["name"]
    if name in DSET_NAMES:
        names = DSET_NAMES
        group = "/PWLR/"
    elif name in INFO_DSET_NAMES:
        names = INFO_DSET_NAMES
        group = "/INFO/"
    elif name in MAPS_DSET_NAMES:
        names = MAPS_DSET_NAMES
        group = "/Maps/"
    else:
        raise KeyError(f"Unsupported name: {key}")

    return names, group


def _harmonize_data(arr):
    if arr.shape[1] == 30:
        # This is specifically for AMSU flags that have not been repeated for the IASI footprints
        return np.repeat(arr, 4, axis=1)
    return arr


def read_geo(fid, key):
    """Read geolocation and related datasets."""
    dsid = GEO_NAMES[key["name"]]
    add_epoch = False
    if "time" in key["name"]:
        days = fid["/L1C/" + dsid["day"]][()]
        msecs = fid["/L1C/" + dsid["msec"]][()]
        data = _form_datetimes(days, msecs)
        add_epoch = True
        dtype = np.float64
    else:
        data = fid["/L1C/" + dsid][()]
        dtype = np.float32
    data = xr.DataArray(da.from_array(data, chunks=CHUNK_SIZE),
                        name=key["name"], dims=["y", "x"]).astype(dtype)

    if add_epoch:
        data.attrs["sensing_time_epoch"] = EPOCH

    return data


def _form_datetimes(days, msecs):
    """Calculate seconds since EPOCH from days and milliseconds for each of IASI scan."""
    all_datetimes = []
    for i in range(days.size):
        day = int(days[i])
        msec = msecs[i]
        scanline_datetimes = []
        for j in range(int(VALUES_PER_SCAN_LINE / 4)):
            usec = 1000 * (j * VIEW_TIME_ADJUSTMENT + msec)
            delta = (dt.timedelta(days=day, microseconds=usec))
            for _k in range(4):
                scanline_datetimes.append(delta.total_seconds())
        all_datetimes.append(scanline_datetimes)

    return np.array(all_datetimes, dtype=np.float64)


class IASIL2CDRNC(NetCDF4FsspecFileHandler):
    """Reader for IASI L2 CDR in NetCDF format.

    Reader for IASI All Sky Temperature and Humidity Profiles - Climate
    Data Record Release 1.1 - Metop-A and -B. Data and documentation are
    available from http://doi.org/10.15770/EUM_SEC_CLM_0063. Data are
    also available from the EUMETSAT Data Store under ID EO:EUM:DAT:0576.
    """

    def get_dataset(self, data_id, ds_info):
        """Obtain dataset."""
        ds = self[data_id["name"]]
        if "scan_lines" in ds.dims:
            ds = ds.rename(scan_lines="y")
        if "pixels" in ds.dims:
            ds = ds.rename(pixels="x")
        if "_FillValue" in ds.attrs and ds.dtype.kind == "f":
            with xr.set_options(keep_attrs=True):
                # have to inverse the logic due to https://github.com/pydata/xarray/issues/7581
                return xr.where(ds != ds.attrs["_FillValue"], ds, np.nan)
        return ds

    def available_datasets(self, configured_datasets=None):
        """Get available datasets based on what's in the file.

        Returns all datasets in the root group.
        """
        yield from super().available_datasets(configured_datasets)
        common = {"file_type": "iasi_l2_cdr_nc", "resolution": 12000}
        for key in self.file_content:
            if "/" in key:  # not a dataset
                continue
            yield (True, {"name": key} | common | self[key].attrs)



# -*- coding: utf-8 -*-
#
# NWCSAF/MSG PP documentation build configuration file, created by
# sphinx-quickstart on Fri Sep 25 16:58:28 2009.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.
"""Sphinx documentation configuration and setup."""

from __future__ import annotations

import datetime as dt
import os
import sys

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.append(os.path.abspath("../../"))
sys.path.append(os.path.abspath(os.path.dirname(__file__)))

from reader_table import generate_reader_table  # noqa: E402

import satpy  # noqa: E402

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# get version using setuptools-scm
release = satpy.__version__
# The full version, including alpha/beta/rc tags.
# for example take major/minor
version = ".".join(release.split(".")[:2])


class Mock(object):  # noqa
    """Mock class for mocking module instances."""

    def __init__(self, *args, **kwargs):
        """Mask any arguments to mock object."""
        self.__version__ = "0.0.0"

    def __call__(self, *args, **kwargs):
        """Mock a function and class object when accessed from mocked module."""
        return Mock()

    @classmethod
    def __getattr__(cls, name):
        """Mock common module attributes used in documentation."""
        if name in ("__file__", "__path__"):
            return "/dev/null"
        elif name[0] == name[0].upper():
            mockType = type(name, (), {})
            mockType.__module__ = __name__
            return mockType
        elif name == "inf":
            return 0
        else:
            return Mock()


# https://github.com/sphinx-doc/sphinx/issues/3920
MOCK_MODULES = ["h5py"]
for mod_name in MOCK_MODULES:
    sys.modules[mod_name] = Mock()  # type: ignore

autodoc_mock_imports = ["cf", "glymur", "h5netcdf", "holoviews", "imageio", "mipp", "netCDF4",
                        "pygac", "pygrib", "pyhdf", "pyninjotiff",
                        "pyorbital", "pyspectral", "rasterio", "trollimage",
                        "zarr"]
autodoc_type_aliases = {
    "ArrayLike": "numpy.typing.ArrayLike",
    "DTypeLike": "numpy.typing.DTypeLike",
}
autodoc_default_options = {
    "special-members": "__init__, __reduce_ex__",
}
nitpick_ignore_regex = [
    ("py:class", r"yaml\.loader\..*Loader"),
    ("py:class", r"numpy\.float32"),
    ("py:class", r"numpy\.uint8"),
    ("py:class", r"numpy\.uint16"),
    ("py:class", r"numpy\.uint32"),
]
autoclass_content = "both"  # append class __init__ docstring to the class docstring

# auto generate reader table from reader config files
with open("reader_table.rst", mode="w") as f:
    f.write(generate_reader_table())

# -- General configuration -----------------------------------------------------

# sphinxcontrib.apidoc was added to sphinx in 8.2.0 as sphinx.etx.apidoc
needs_sphinx = "8.2.0"

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ["sphinx.ext.autodoc", "sphinx.ext.intersphinx", "sphinx.ext.todo", "sphinx.ext.coverage",
              "sphinx.ext.doctest", "sphinx.ext.napoleon", "sphinx.ext.autosummary", "sphinx.ext.autosectionlabel",
              "doi_role", "sphinx.ext.viewcode", "sphinx.ext.apidoc",
              "sphinx.ext.mathjax", "sphinx_autodoc_typehints"]

linkcheck_allowed_redirects: dict[str, str] = {}

# Autosectionlabel
# Make sure target is unique
autosectionlabel_prefix_document = True
autosectionlabel_maxdepth = 3

# API docs
apidoc_modules = [
    {
        "path": "../../satpy",
        "destination": "api/",
        "exclude_patterns": [
            "../../satpy/readers/caliop_l2_cloud.py",
            "../../satpy/readers/ghrsst_l3c_sst.py",
            "../../satpy/readers/scatsat1_l2b.py",
            # Prefer to not document test modules. Most users will look at
            # source code if needed and we want to avoid documentation builds
            # suffering from import-time test data creation. We want to keep
            # things contributors might be interested in like satpy.tests.utils.
            "../../satpy/tests/test_*.py",
            "../../satpy/tests/**/test_*.py",
            "../../satpy/tests/*_tests/*",
        ],
    },
]
apidoc_separate_modules = True
apidoc_include_private = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = u"Satpy"
copyright = u"2009-{}, The PyTroll Team".format(dt.datetime.now(dt.timezone.utc).strftime("%Y"))  # noqa: A001

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of documents that shouldn't be included in the build.
# unused_docs = []

# List of directories, relative to source directory, that shouldn't be searched
# for source files.
exclude_trees: list[str] = []

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_css_files = [
    "theme_overrides.css",  # override wide tables in RTD theme
    # "https://cdn.datatables.net/v/dt/dt-2.0.0/datatables.min.css",
    "https://cdn.datatables.net/v/dt/dt-2.0.8/r-3.0.2/datatables.min.css"
]

html_js_files = [
    # "https://cdn.datatables.net/v/dt/dt-2.0.0/datatables.min.js",
    "https://cdn.datatables.net/v/dt/dt-2.0.8/r-3.0.2/datatables.min.js",
    "main.js",
]


# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_use_modindex = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = ''

# Output file base name for HTML help builder.
htmlhelp_basename = "NWCSAFMSGPPdoc"


# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  ("index", "satpy.tex", "Satpy Documentation",
   "Satpy Developers", "manual"),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_use_modindex = True


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "dask": ("https://docs.dask.org/en/latest", None),
    "geoviews": ("https://geoviews.org", None),
    "jobqueue": ("https://jobqueue.dask.org/en/latest", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "pydecorate": ("https://pydecorate.readthedocs.io/en/stable", None),
    "pyorbital": ("https://pyorbital.readthedocs.io/en/stable", None),
    "pyproj": ("https://pyproj4.github.io/pyproj/dev", None),
    "pyresample": ("https://pyresample.readthedocs.io/en/stable", None),
    "pytest": ("https://docs.pytest.org/en/stable/", None),
    "python": ("https://docs.python.org/3", None),
    "scipy": ("https://scipy.github.io/devdocs", None),
    "trollimage": ("https://trollimage.readthedocs.io/en/stable", None),
    "trollsift": ("https://trollsift.readthedocs.io/en/stable", None),
    "xarray": ("https://docs.xarray.dev/en/stable", None),
    "rasterio": ("https://rasterio.readthedocs.io/en/latest", None),
    "donfig": ("https://donfig.readthedocs.io/en/latest", None),
    "pooch": ("https://www.fatiando.org/pooch/latest/", None),
    "fsspec": ("https://filesystem-spec.readthedocs.io/en/latest/", None),
    "asv": ("https://asv.readthedocs.io/en/latest", None),
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019, 2022, 2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.viirs_sdr module."""

import os
import unittest
from contextlib import contextmanager
from unittest import mock

import numpy as np
import pytest

from satpy.readers.core.viirs_atms_sdr import DATASET_KEYS
from satpy.tests.reader_tests.test_hdf5_utils import FakeHDF5FileHandler

DEFAULT_FILE_DTYPE = np.uint16
DEFAULT_FILE_SHAPE = (32, 300)
# Mimicking one scan line of data
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)
DEFAULT_FILE_FACTORS = np.array([2.0, 1.0], dtype=np.float32)


class FakeHDF5FileHandler2(FakeHDF5FileHandler):
    """Swap-in HDF5 File Handler."""

    _num_test_granules = 1
    _num_scans_per_gran = [48]

    def __init__(self, filename, filename_info, filetype_info, include_factors=True):
        """Create fake file handler."""
        self.include_factors = include_factors
        super(FakeHDF5FileHandler2, self).__init__(filename, filename_info, filetype_info)

    @staticmethod
    def _add_basic_metadata_to_file_content(file_content, filename_info, num_grans):
        start_time = filename_info["start_time"]
        end_time = filename_info["end_time"].replace(year=start_time.year,
                                                     month=start_time.month,
                                                     day=start_time.day)
        begin_date = start_time.strftime("%Y%m%d")
        begin_date = np.array(begin_date)

        begin_time = start_time.strftime("%H%M%S.%fZ")
        begin_time = np.array(begin_time)

        ending_date = end_time.strftime("%Y%m%d")
        ending_time = end_time.strftime("%H%M%S.%fZ")
        new_file_content = {
            "{prefix2}/attr/AggregateNumberGranules": num_grans,
            "{prefix2}/attr/AggregateBeginningDate": begin_date,
            "{prefix2}/attr/AggregateBeginningTime": begin_time,
            "{prefix2}/attr/AggregateEndingDate": ending_date,
            "{prefix2}/attr/AggregateEndingTime": ending_time,
            "{prefix2}/attr/G-Ring_Longitude": np.array([0.0, 0.1, 0.2, 0.3]),
            "{prefix2}/attr/G-Ring_Latitude": np.array([0.0, 0.1, 0.2, 0.3]),
            "{prefix2}/attr/AggregateBeginningOrbitNumber": "{0:d}".format(filename_info["orbit"]),
            "{prefix2}/attr/AggregateEndingOrbitNumber": "{0:d}".format(filename_info["orbit"]),
            "{prefix1}/attr/Instrument_Short_Name": "VIIRS",
            "/attr/Platform_Short_Name": "NPP",
        }
        file_content.update(new_file_content)

    def _add_granule_specific_info_to_file_content(
            self,
            file_content, dataset_group, num_granules, num_scans_per_granule,
            gran_group_prefix):
        lons_lists = self._get_per_granule_lons()
        lats_lists = self._get_per_granule_lats()
        file_content["{prefix3}/NumberOfScans"] = np.array([48] * num_granules)
        for granule_idx in range(num_granules):
            prefix_gran = "{prefix}/{dataset_group}_Gran_{idx}".format(prefix=gran_group_prefix,
                                                                       dataset_group=dataset_group,
                                                                       idx=granule_idx)
            num_scans = num_scans_per_granule[granule_idx]
            file_content[prefix_gran + "/attr/N_Number_Of_Scans"] = num_scans
            file_content[prefix_gran + "/attr/G-Ring_Longitude"] = lons_lists[granule_idx]
            file_content[prefix_gran + "/attr/G-Ring_Latitude"] = lats_lists[granule_idx]

    @staticmethod
    def _get_per_granule_lons():
        return [
            np.array(
                [
                    50.51393, 49.566296, 48.865967, 18.96082, -4.0238385,
                    -7.05221, -10.405702, 14.638646
                ],
                dtype=np.float32),
            np.array(
                [
                    53.52594, 51.685738, 50.439102, 14.629087, -10.247547,
                    -13.951393, -18.256989, 8.36572
                ],
                dtype=np.float32),
            np.array(
                [
                    59.386833, 55.770416, 53.38952, 8.353765, -18.062435,
                    -22.608992, -27.867302, -1.3537619
                ],
                dtype=np.float32),
            np.array(
                [
                    72.50243, 64.17125, 59.15234, -1.3654504, -27.620953,
                    -33.091743, -39.28113, -17.749891
                ],
                dtype=np.float32)
        ]

    @staticmethod
    def _get_per_granule_lats():
        return [
            np.array(
                [
                    67.969505, 65.545685, 63.103046, 61.853905, 55.169273,
                    57.062447, 58.86063, 66.495514
                ],
                dtype=np.float32),
            np.array(
                [
                    72.74879, 70.2493, 67.84738, 66.49691, 58.77254,
                    60.465942, 62.11525, 71.08249
                ],
                dtype=np.float32),
            np.array(
                [
                    77.393425, 74.977875, 72.62976, 71.083435, 62.036346,
                    63.465122, 64.78075, 75.36842
                ],
                dtype=np.float32),
            np.array(
                [
                    81.67615, 79.49934, 77.278656, 75.369415, 64.72178,
                    65.78417, 66.66166, 79.00025
                ],
                dtype=np.float32),
        ]

    def _add_data_info_to_file_content(self, file_content, filename, data_var_prefix, num_grans):
        # SDR files always produce data with 48 scans per granule even if there are less
        total_rows = DEFAULT_FILE_SHAPE[0] * 48 * num_grans
        new_shape = (total_rows, DEFAULT_FILE_SHAPE[1])
        if filename[2:5] in ["M{:02d}".format(x) for x in range(12)] + ["I01", "I02", "I03"]:
            keys = ["Radiance", "Reflectance"]
        elif filename[2:5] in ["M{:02d}".format(x) for x in range(12, 17)] + ["I04", "I05"]:
            keys = ["Radiance", "BrightnessTemperature"]
        else:
            # DNB
            keys = ["Radiance"]

        for k in keys:
            k = data_var_prefix + "/" + k
            file_content[k] = np.repeat(DEFAULT_FILE_DATA.copy(), 48 * num_grans, axis=0)
            file_content[k + "/shape"] = new_shape
            if self.include_factors:
                file_content[k + "Factors"] = np.repeat(
                    DEFAULT_FILE_FACTORS.copy()[None, :], num_grans, axis=0).ravel()

    @staticmethod
    def _add_geolocation_info_to_file_content(file_content, filename, data_var_prefix, num_grans):
        # SDR files always produce data with 48 scans per granule even if there are less
        total_rows = DEFAULT_FILE_SHAPE[0] * 48 * num_grans
        new_shape = (total_rows, DEFAULT_FILE_SHAPE[1])
        is_dnb = filename[:5] not in ["GMODO", "GIMGO"]
        if not is_dnb:
            lon_data = np.linspace(15, 55, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
            lat_data = np.linspace(55, 75, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
        else:
            lon_data = np.linspace(5, 45, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
            lat_data = np.linspace(45, 65, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)

        for k in ["Latitude"]:
            k = data_var_prefix + "/" + k
            file_content[k] = lat_data
            file_content[k] = np.repeat([file_content[k]], total_rows, axis=0)
            file_content[k + "/shape"] = new_shape
        for k in ["Longitude"]:
            k = data_var_prefix + "/" + k
            file_content[k] = lon_data
            file_content[k] = np.repeat([file_content[k]], total_rows, axis=0)
            file_content[k + "/shape"] = new_shape

        angles = ["SolarZenithAngle",
                  "SolarAzimuthAngle",
                  "SatelliteZenithAngle",
                  "SatelliteAzimuthAngle"]
        if is_dnb:
            angles += ["LunarZenithAngle", "LunarAzimuthAngle"]
        for k in angles:
            k = data_var_prefix + "/" + k
            file_content[k] = lon_data  # close enough to SZA
            file_content[k] = np.repeat([file_content[k]], total_rows, axis=0)
            file_content[k + "/shape"] = new_shape

    @staticmethod
    def _add_geo_ref(file_content, filename):
        if filename[:3] == "SVI":
            geo_prefix = "GIMGO"
        elif filename[:3] == "SVM":
            geo_prefix = "GMODO"
        else:
            geo_prefix = None
        if geo_prefix:
            file_content["/attr/N_GEO_Ref"] = geo_prefix + filename[5:]

    @staticmethod
    def _convert_numpy_content_to_dataarray(final_content):
        import dask.array as da
        from xarray import DataArray
        for key, val in final_content.items():
            if isinstance(val, np.ndarray):
                val = da.from_array(val, chunks=val.shape)
                if val.ndim > 1:
                    final_content[key] = DataArray(val, dims=("y", "x"))
                else:
                    final_content[key] = DataArray(val)

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        final_content = {}
        for dataset in self.datasets:
            dataset_group = DATASET_KEYS[dataset]
            prefix1 = "Data_Products/{dataset_group}".format(dataset_group=dataset_group)
            prefix2 = "{prefix}/{dataset_group}_Aggr".format(prefix=prefix1, dataset_group=dataset_group)
            prefix3 = "All_Data/{dataset_group}_All".format(dataset_group=dataset_group)

            file_content = {}
            self._add_basic_metadata_to_file_content(file_content, filename_info, self._num_test_granules)
            self._add_granule_specific_info_to_file_content(file_content, dataset_group,
                                                            self._num_test_granules, self._num_scans_per_gran,
                                                            prefix1)
            self._add_geo_ref(file_content, filename)

            for k, v in list(file_content.items()):
                file_content[k.format(prefix1=prefix1, prefix2=prefix2, prefix3=prefix3)] = v

            if filename[:3] in ["SVM", "SVI", "SVD"]:
                self._add_data_info_to_file_content(file_content, filename, prefix3,
                                                    self._num_test_granules)
            elif filename[0] == "G":
                self._add_geolocation_info_to_file_content(file_content, filename, prefix3,
                                                           self._num_test_granules)
            final_content.update(file_content)
        self._convert_numpy_content_to_dataarray(final_content)
        return final_content


@contextmanager
def touch_geo_files(*prefixes):
    """Create and then remove VIIRS SDR geolocation files."""
    geofiles = [_touch_geo_file(prefix) for prefix in prefixes]
    try:
        yield geofiles
    finally:
        for filename in geofiles:
            os.remove(filename)


def _touch_geo_file(prefix):
    geo_fn = prefix + "_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5"
    open(geo_fn, "w")
    return geo_fn


class TestVIIRSSDRReader(unittest.TestCase):
    """Test VIIRS SDR Reader."""

    yaml_file = "viirs_sdr.yaml"

    def _assert_reflectance_properties(self, data_arr, num_scans=16, with_area=True):
        assert np.issubdtype(data_arr.dtype, np.float32)
        assert data_arr.attrs["calibration"] == "reflectance"
        assert data_arr.attrs["units"] == "%"
        assert data_arr.attrs["rows_per_scan"] == num_scans
        if with_area:
            assert "area" in data_arr.attrs
            assert data_arr.attrs["area"] is not None
            assert data_arr.attrs["area"].shape == data_arr.shape
        else:
            assert "area" not in data_arr.attrs

    def _assert_bt_properties(self, data_arr, num_scans=16, with_area=True):
        assert np.issubdtype(data_arr.dtype, np.float32)
        assert data_arr.attrs["calibration"] == "brightness_temperature"
        assert data_arr.attrs["units"] == "K"
        assert data_arr.attrs["rows_per_scan"] == num_scans
        if with_area:
            assert "area" in data_arr.attrs
            assert data_arr.attrs["area"] is not None
            assert data_arr.attrs["area"].shape == data_arr.shape
        else:
            assert "area" not in data_arr.attrs

    def _assert_dnb_radiance_properties(self, data_arr, with_area=True):
        assert np.issubdtype(data_arr.dtype, np.float32)
        assert data_arr.attrs["calibration"] == "radiance"
        assert data_arr.attrs["units"] == "W m-2 sr-1"
        assert data_arr.attrs["rows_per_scan"] == 16
        if with_area:
            assert "area" in data_arr.attrs
            assert data_arr.attrs["area"] is not None
            assert data_arr.attrs["area"].shape == data_arr.shape
        else:
            assert "area" not in data_arr.attrs

    def setUp(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.core.viirs_atms_sdr import JPSS_SDR_FileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(JPSS_SDR_FileHandler, "__bases__", (FakeHDF5FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    def test_init_start_time_is_nodate(self):
        """Test basic init with start_time being set to the no-date 1/1-1958."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with pytest.raises(ValueError, match="Datetime invalid 1958-01-01 00:00:00"):
            _ = r.create_filehandlers([
                "SVI01_npp_d19580101_t0000000_e0001261_b01708_c20120226002130255476_noaa_ops.h5",
            ])

    def test_init_start_time_beyond(self):
        """Test basic init with start_time after the provided files."""
        import datetime as dt

        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs,
                        filter_parameters={
                            "start_time": dt.datetime(2012, 2, 26)
                        })
        fhs = r.create_filehandlers([
            "SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        assert len(fhs) == 0

    def test_init_end_time_beyond(self):
        """Test basic init with end_time before the provided files."""
        import datetime as dt

        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs,
                        filter_parameters={
                            "end_time": dt.datetime(2012, 2, 24)
                        })
        fhs = r.create_filehandlers([
            "SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        assert len(fhs) == 0

    def test_init_start_end_time(self):
        """Test basic init with end_time before the provided files."""
        import datetime as dt

        from satpy.readers.core.loading import load_reader

        r = load_reader(self.reader_configs,
                        filter_parameters={
                            "start_time": dt.datetime(2012, 2, 24),
                            "end_time": dt.datetime(2012, 2, 26)
                        })
        loadables = r.select_files_from_pathnames([
            "SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    def test_load_all_m_reflectances_no_geo(self):
        """Load all M band reflectances with no geo files provided."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "SVM01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM02_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM03_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM04_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM05_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM06_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM07_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM08_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM09_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM10_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM11_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        r.create_filehandlers(loadables)
        ds = r.load(["M01",
                     "M02",
                     "M03",
                     "M04",
                     "M05",
                     "M06",
                     "M07",
                     "M08",
                     "M09",
                     "M10",
                     "M11",
                     ])
        assert len(ds) == 11
        for d in ds.values():
            self._assert_reflectance_properties(d, with_area=False)

    def test_load_all_m_reflectances_find_geo(self):
        """Load all M band reflectances with geo files not specified but existing."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "SVM01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM02_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM03_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM04_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM05_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM06_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM07_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM08_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM09_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM10_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM11_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        with touch_geo_files("GMTCO", "GMODO") as (geo_fn1, geo_fn2):
            r.create_filehandlers(loadables)
            ds = r.load(["M01",
                         "M02",
                         "M03",
                         "M04",
                         "M05",
                         "M06",
                         "M07",
                         "M08",
                         "M09",
                         "M10",
                         "M11",
                         ])

        assert len(ds) == 11
        for d in ds.values():
            self._assert_reflectance_properties(d, with_area=True)

    def test_load_all_m_reflectances_provided_geo(self):
        """Load all M band reflectances with geo files provided."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "SVM01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM02_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM03_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM04_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM05_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM06_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM07_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM08_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM09_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM10_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM11_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "GMTCO_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        with touch_geo_files("GMTCO", "GMODO") as (geo_fn1, geo_fn2):
            r.create_filehandlers(loadables)
            ds = r.load(["M01",
                         "M02",
                         "M03",
                         "M04",
                         "M05",
                         "M06",
                         "M07",
                         "M08",
                         "M09",
                         "M10",
                         "M11",
                         ])
        assert len(ds) == 11
        for d in ds.values():
            self._assert_reflectance_properties(d, with_area=True)
            assert d.attrs["area"].lons.min() == 5
            assert d.attrs["area"].lats.min() == 45
            assert d.attrs["area"].lons.attrs["rows_per_scan"] == 16
            assert d.attrs["area"].lats.attrs["rows_per_scan"] == 16

    def test_load_all_m_reflectances_use_nontc(self):
        """Load all M band reflectances but use non-TC geolocation."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs, use_tc=False)
        loadables = r.select_files_from_pathnames([
            "SVM01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM02_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM03_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM04_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM05_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM06_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM07_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM08_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM09_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM10_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM11_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "GMTCO_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "GMODO_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        with touch_geo_files("GMTCO", "GMODO") as (geo_fn1, geo_fn2):
            r.create_filehandlers(loadables, {"use_tc": False})
            ds = r.load(["M01",
                         "M02",
                         "M03",
                         "M04",
                         "M05",
                         "M06",
                         "M07",
                         "M08",
                         "M09",
                         "M10",
                         "M11",
                         ])
        assert len(ds) == 11
        for d in ds.values():
            self._assert_reflectance_properties(d, with_area=True)
            assert d.attrs["area"].lons.min() == 15
            assert d.attrs["area"].lats.min() == 55
            assert d.attrs["area"].lons.attrs["rows_per_scan"] == 16
            assert d.attrs["area"].lats.attrs["rows_per_scan"] == 16

    def test_load_all_m_reflectances_use_nontc2(self):
        """Load all M band reflectances but use non-TC geolocation because TC isn't available."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs, use_tc=None)
        loadables = r.select_files_from_pathnames([
            "SVM01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM02_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM03_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM04_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM05_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM06_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM07_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM08_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM09_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM10_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM11_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "GMODO_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        with touch_geo_files("GMODO") as (geo_fn2,):
            r.create_filehandlers(loadables, {"use_tc": None})
            ds = r.load(["M01",
                         "M02",
                         "M03",
                         "M04",
                         "M05",
                         "M06",
                         "M07",
                         "M08",
                         "M09",
                         "M10",
                         "M11",
                         ])
        assert len(ds) == 11
        for d in ds.values():
            self._assert_reflectance_properties(d, with_area=True)
            assert d.attrs["area"].lons.min() == 15
            assert d.attrs["area"].lats.min() == 55
            assert d.attrs["area"].lons.attrs["rows_per_scan"] == 16
            assert d.attrs["area"].lats.attrs["rows_per_scan"] == 16

    def test_load_all_m_bts(self):
        """Load all M band brightness temperatures."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "SVM12_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM13_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM14_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM15_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM16_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "GMTCO_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        r.create_filehandlers(loadables)
        ds = r.load(["M12",
                     "M13",
                     "M14",
                     "M15",
                     "M16",
                     ])
        assert len(ds) == 5
        for d in ds.values():
            self._assert_bt_properties(d, with_area=True)

    def test_load_dnb_sza_no_factors(self):
        """Load DNB solar zenith angle with no scaling factors.

        The angles in VIIRS SDRs should never have scaling factors so we test
        it that way.

        """
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "GDNBO_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        r.create_filehandlers(loadables, {"include_factors": False})
        ds = r.load(["dnb_solar_zenith_angle",
                     "dnb_solar_azimuth_angle",
                     "dnb_satellite_zenith_angle",
                     "dnb_satellite_azimuth_angle",
                     "dnb_lunar_zenith_angle",
                     "dnb_lunar_azimuth_angle"])
        assert len(ds) == 6
        for d in ds.values():
            assert np.issubdtype(d.dtype, np.float32)
            assert d.attrs["units"] == "degrees"
            assert d.attrs["rows_per_scan"] == 16
            assert "area" in d.attrs
            assert d.attrs["area"] is not None

    def test_load_all_m_radiances(self):
        """Load all M band radiances."""
        from satpy.readers.core.loading import load_reader
        from satpy.tests.utils import make_dsq
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "SVM01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM02_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM03_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM04_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM05_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM06_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM07_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM08_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM09_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM10_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM11_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM12_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM13_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM14_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM15_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVM16_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "GMTCO_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        r.create_filehandlers(loadables)
        ds = r.load([
            make_dsq(name="M01", calibration="radiance"),
            make_dsq(name="M02", calibration="radiance"),
            make_dsq(name="M03", calibration="radiance"),
            make_dsq(name="M04", calibration="radiance"),
            make_dsq(name="M05", calibration="radiance"),
            make_dsq(name="M06", calibration="radiance"),
            make_dsq(name="M07", calibration="radiance"),
            make_dsq(name="M08", calibration="radiance"),
            make_dsq(name="M09", calibration="radiance"),
            make_dsq(name="M10", calibration="radiance"),
            make_dsq(name="M11", calibration="radiance"),
            make_dsq(name="M12", calibration="radiance"),
            make_dsq(name="M13", calibration="radiance"),
            make_dsq(name="M14", calibration="radiance"),
            make_dsq(name="M15", calibration="radiance"),
            make_dsq(name="M16", calibration="radiance"),
        ])
        assert len(ds) == 16
        for d in ds.values():
            assert np.issubdtype(d.dtype, np.float32)
            assert d.attrs["calibration"] == "radiance"
            assert d.attrs["units"] == "W m-2 um-1 sr-1"
            assert d.attrs["rows_per_scan"] == 16
            assert "area" in d.attrs
            assert d.attrs["area"] is not None

    def test_load_dnb(self):
        """Load DNB dataset."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "SVDNB_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "GDNBO_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        r.create_filehandlers(loadables)
        ds = r.load(["DNB"])
        assert len(ds) == 1
        for d in ds.values():
            data = d.values
            # default scale factors are 2 and offset 1
            # multiply DNB by 10000 should mean the first value of 0 should be:
            # data * factor * 10000 + offset * 10000
            # 0 * 2 * 10000 + 1 * 10000 => 10000
            assert data[0, 0] == 10000
            # the second value of 1 should be:
            # 1 * 2 * 10000 + 1 * 10000 => 30000
            assert data[0, 1] == 30000
            self._assert_dnb_radiance_properties(d, with_area=True)

    def test_load_dnb_no_factors(self):
        """Load DNB dataset with no provided scale factors."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "SVDNB_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "GDNBO_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        r.create_filehandlers(loadables, {"include_factors": False})
        ds = r.load(["DNB"])
        assert len(ds) == 1
        for d in ds.values():
            data = d.values
            # no scale factors, default factor 1 and offset 0
            # multiply DNB by 10000 should mean the first value of 0 should be:
            # data * factor * 10000 + offset * 10000
            # 0 * 1 * 10000 + 0 * 10000 => 0
            assert data[0, 0] == 0
            # the second value of 1 should be:
            # 1 * 1 * 10000 + 0 * 10000 => 10000
            assert data[0, 1] == 10000
            self._assert_dnb_radiance_properties(d, with_area=True)

    def test_load_i_no_files(self):
        """Load I01 when only DNB files are provided."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "SVDNB_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "GDNBO_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        r.create_filehandlers(loadables)
        assert "I01" not in [x["name"] for x in r.available_dataset_ids]
        ds = r.load(["I01"])
        assert len(ds) == 0

    def test_load_all_i_reflectances_provided_geo(self):
        """Load all I band reflectances with geo files provided."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVI02_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVI03_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "GITCO_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        r.create_filehandlers(loadables)
        ds = r.load(["I01",
                     "I02",
                     "I03",
                     ])
        assert len(ds) == 3
        for d in ds.values():
            self._assert_reflectance_properties(d, num_scans=32)
            assert d.attrs["area"].lons.min() == 5
            assert d.attrs["area"].lats.min() == 45
            assert d.attrs["area"].lons.attrs["rows_per_scan"] == 32
            assert d.attrs["area"].lats.attrs["rows_per_scan"] == 32

    def test_load_all_i_bts(self):
        """Load all I band brightness temperatures."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "SVI04_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVI05_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "GITCO_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        r.create_filehandlers(loadables)
        ds = r.load(["I04",
                     "I05",
                     ])
        assert len(ds) == 2
        for d in ds.values():
            self._assert_bt_properties(d, num_scans=32)

    def test_load_all_i_radiances(self):
        """Load all I band radiances."""
        from satpy.readers.core.loading import load_reader
        from satpy.tests.utils import make_dsq
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVI02_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVI03_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVI04_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "SVI05_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
            "GITCO_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        r.create_filehandlers(loadables)
        ds = r.load([
            make_dsq(name="I01", calibration="radiance"),
            make_dsq(name="I02", calibration="radiance"),
            make_dsq(name="I03", calibration="radiance"),
            make_dsq(name="I04", calibration="radiance"),
            make_dsq(name="I05", calibration="radiance"),
        ])
        assert len(ds) == 5
        for d in ds.values():
            assert np.issubdtype(d.dtype, np.float32) is True
            assert d.attrs["calibration"] == "radiance"
            assert d.attrs["units"] == "W m-2 um-1 sr-1"
            assert d.attrs["rows_per_scan"] == 32
            assert "area" in d.attrs
            assert d.attrs["area"] is not None


class FakeHDF5FileHandlerAggr(FakeHDF5FileHandler2):
    """Swap-in HDF5 File Handler with 4 VIIRS Granules per file."""

    _num_test_granules = 4
    _num_scans_per_gran = [48] * 4


class TestAggrVIIRSSDRReader(unittest.TestCase):
    """Test VIIRS SDR Reader."""

    yaml_file = "viirs_sdr.yaml"

    def setUp(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.viirs_sdr import VIIRSSDRFileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(VIIRSSDRFileHandler, "__bases__", (FakeHDF5FileHandlerAggr,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()

    def test_bounding_box(self):
        """Test bounding box."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        r.create_filehandlers(loadables)
        # make sure we have some files
        expected_lons = [
            72.50243, 64.17125, 59.15234, 59.386833, 55.770416, 53.38952, 53.52594, 51.685738, 50.439102, 50.51393,
            49.566296, 48.865967, 18.96082, -4.0238385, -7.05221, -10.247547, -13.951393, -18.062435, -22.608992,
            -27.620953, -33.091743, -39.28113, -17.749891
        ]
        expected_lats = [
            81.67615, 79.49934, 77.278656, 77.393425, 74.977875, 72.62976, 72.74879, 70.2493, 67.84738, 67.969505,
            65.545685, 63.103046, 61.853905, 55.169273, 57.062447, 58.77254, 60.465942, 62.036346, 63.465122,
            64.72178, 65.78417, 66.66166, 79.00025
        ]
        lons, lats = r.file_handlers["generic_file"][0].get_bounding_box()
        np.testing.assert_allclose(lons, expected_lons)
        np.testing.assert_allclose(lats, expected_lats)


class FakeShortHDF5FileHandlerAggr(FakeHDF5FileHandler2):
    """Fake file that has less scans than usual in a couple granules."""

    _num_test_granules = 3
    _num_scans_per_gran = [47, 48, 47]


class TestShortAggrVIIRSSDRReader(unittest.TestCase):
    """Test VIIRS SDR Reader with a file that has truncated granules."""

    yaml_file = "viirs_sdr.yaml"

    def setUp(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.core.viirs_atms_sdr import JPSS_SDR_FileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(JPSS_SDR_FileHandler, "__bases__", (FakeShortHDF5FileHandlerAggr,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()

    def test_load_truncated_band(self):
        """Test loading a single truncated band."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5",
        ])
        r.create_filehandlers(loadables)
        ds = r.load(["I01"])
        assert len(ds) == 1
        i01_data = ds["I01"].compute()
        expected_rows = sum(FakeShortHDF5FileHandlerAggr._num_scans_per_gran) * DEFAULT_FILE_SHAPE[0]
        assert i01_data.shape == (expected_rows, 300)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the Vaisala GLD360 UALF2-reader."""

import datetime as dt

import numpy as np
import pytest

from satpy.readers.gld360_ualf2 import UALF2_COLUMN_NAMES, VaisalaGld360Ualf2FileHandler
from satpy.tests.utils import make_dataid

TEST_START_TIME = dt.datetime(2021, 1, 4, 8, 0)
TEST_END_TIME = TEST_START_TIME + dt.timedelta(hours=1)


@pytest.fixture
def fake_file(tmp_path):
    """Create UALF2 file for the tests."""
    fname = tmp_path / "2021.01.04.08.00.txt"
    with open(fname, "w", encoding="utf-8") as fid:
        fid.write(
            "2\t3\t2021\t1\t4\t8\t0\t1\t51\t-20.8001\t-158.3439\t0\t0\t10\t0\t0\t1\t3\t3\t9.47\t"
            "1.91\t1.59\t0.19\t11.4\t8.8\t0.0\t1\t1\t0\t1\n"
            "2\t3\t2021\t1\t4\t8\t0\t1\t864782486\t0.4381\t-0.8500\t0\t0\t-20\t0\t1\t0\t4\t5\t"
            "24.99\t1.95\t1.53\t1.53\t14.0\t12.9\t-0.0\t0\t1\t0\t1\n"
            "2\t3\t2021\t1\t4\t8\t0\t1\t864782486\t0.4381\t-0.8500\t0\t0\t-20\t0\t1\t0\t4\t5\t"
            "24.99\t1.95\t1.53\t1.53\t14.0\t12.9\t-0.0\t0\t1\t0\t1\n"
            "2\t3\t2021\t1\t4\t8\t0\t1\t897014133\t66.8166\t42.4914\t0\t0\t15\t0\t0\t1\t5\t7\t"
            "103.87\t4.33\t1.46\t0.48\t22.0\t12.3\t0.0\t1\t1\t0\t1"
        )

    return fname


@pytest.fixture
def fake_filehandler(fake_file):
    """Create FileHandler for the tests."""
    filename_info = {}
    filetype_info = {}

    return VaisalaGld360Ualf2FileHandler(
        fake_file, filename_info, filetype_info
    )


def test_ualf2_record_type(fake_filehandler):
    """Test UALF2 record type."""
    expected = np.array([2, 2, 2])
    dataset_id = make_dataid(name="ualf_record_type")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_array_equal(actual, expected)


def test_network_type(fake_filehandler):
    """Test network type."""
    expected = np.array([3, 3, 3])
    dataset_id = make_dataid(name="network_type")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_array_equal(actual, expected)


def test_time(fake_filehandler):
    """Test time."""
    expected = np.array(
        [
            "2021-01-04T08:00:01.000000051",
            "2021-01-04T08:00:01.864782486",
            "2021-01-04T08:00:01.897014133",
        ],
        dtype="datetime64[ns]",
    )
    dataset_id = make_dataid(name="time")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_array_equal(actual, expected)


def test_latitude(fake_filehandler):
    """Test latitude."""
    expected = np.array([-20.8001, 0.4381, 66.8166])
    dataset_id = make_dataid(name="latitude")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_allclose(actual, expected, rtol=1e-05)


def test_longitude(fake_filehandler):
    """Test longitude."""
    expected = np.array([-158.3439, -0.85, 42.4914])
    dataset_id = make_dataid(name="longitude")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_allclose(actual, expected, rtol=1e-05)


def test_altitude(fake_filehandler):
    """Test altitude."""
    expected = np.array([0, 0, 0])
    dataset_id = make_dataid(name="altitude")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_array_equal(actual, expected)


def test_altitude_uncertainty(fake_filehandler):
    """Test altitude uncertainty."""
    expected = np.array([0, 0, 0])
    dataset_id = make_dataid(name="altitude_uncertainty")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_array_equal(actual, expected)


def test_peak_current(fake_filehandler):
    """Test peak current."""
    expected = np.array([10, -20, 15])
    dataset_id = make_dataid(name="peak_current")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_array_equal(actual, expected)


def test_vhf_range(fake_filehandler):
    """Test vhf range."""
    expected = np.array([0, 0, 0])
    dataset_id = make_dataid(name="vhf_range")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_array_equal(actual, expected)


def test_multiplicity_flash(fake_filehandler):
    """Test multiplicity flash."""
    expected = np.array([0, 1, 0])
    dataset_id = make_dataid(name="multiplicity_flash")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_array_equal(actual, expected)


def test_cloud_pulse_count(fake_filehandler):
    """Test cloud pulse count."""
    expected = np.array([1, 0, 1])
    dataset_id = make_dataid(name="cloud_pulse_count")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_array_equal(actual, expected)


def test_number_of_sensors(fake_filehandler):
    """Test number of sensors."""
    expected = np.array([3, 4, 5])
    dataset_id = make_dataid(name="number_of_sensors")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_array_equal(actual, expected)


def test_degree_freedom_for_location(fake_filehandler):
    """Test degree freedom for location."""
    expected = np.array([3, 5, 7])
    dataset_id = make_dataid(name="degree_freedom_for_location")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_array_equal(actual, expected)


def test_error_ellipse_angle(fake_filehandler):
    """Test error ellipse angle."""
    expected = np.array([9.47, 24.99, 103.87])
    dataset_id = make_dataid(name="error_ellipse_angle")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_allclose(actual, expected, rtol=1e-05)


def test_error_ellipse_max_axis_length(fake_filehandler):
    """Test error ellipse max axis length."""
    expected = np.array([1.91, 1.95, 4.33])
    dataset_id = make_dataid(name="error_ellipse_max_axis_length")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_allclose(actual, expected, rtol=1e-05)


def test_error_ellipse_min_axis_length(fake_filehandler):
    """Test error ellipse min axis length."""
    expected = np.array([1.59, 1.53, 1.46])
    dataset_id = make_dataid(name="error_ellipse_min_axis_length")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_allclose(actual, expected, rtol=1e-05)


def test_chi_squared_value_location_optimization(fake_filehandler):
    """Test chi squared value location optimization."""
    expected = np.array([0.19, 1.53, 0.48])
    dataset_id = make_dataid(name="chi_squared_value_location_optimization")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_allclose(actual, expected, rtol=1e-05)


def test_wave_form_rise_time(fake_filehandler):
    """Test wave form rise time."""
    expected = np.array([11.4, 14.0, 22.0])
    dataset_id = make_dataid(name="wave_form_rise_time")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_allclose(actual, expected, rtol=1e-05)


def test_wave_form_peak_to_zero_time(fake_filehandler):
    """Test wave form peak to zero time."""
    expected = np.array([8.8, 12.9, 12.3])
    dataset_id = make_dataid(name="wave_form_peak_to_zero_time")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_allclose(actual, expected, rtol=1e-05)


def test_wave_form_max_rate_of_rise(fake_filehandler):
    """Test wave form max rate of rise."""
    expected = np.array([0, 0, 0])
    dataset_id = make_dataid(name="wave_form_max_rate_of_rise")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_array_equal(actual, expected)


def test_cloud_indicator(fake_filehandler):
    """Test cloud indicator."""
    expected = np.array([1, 0, 1])
    dataset_id = make_dataid(name="cloud_indicator")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_array_equal(actual, expected)


def test_angle_indicator(fake_filehandler):
    """Test angle indicator."""
    expected = np.array([1, 1, 1])
    dataset_id = make_dataid(name="angle_indicator")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_array_equal(actual, expected)


def test_signal_indicator(fake_filehandler):
    """Test signal indicator."""
    expected = np.array([0, 0, 0])
    dataset_id = make_dataid(name="signal_indicator")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_array_equal(actual, expected)


def test_timing_indicator(fake_filehandler):
    """Test timing indicator."""
    expected = np.array([1, 1, 1])
    dataset_id = make_dataid(name="timing_indicator")
    dataset_info = {}
    actual = fake_filehandler.get_dataset(dataset_id, dataset_info).values
    np.testing.assert_array_equal(actual, expected)


def test_pad_nanoseconds(fake_filehandler):
    """Test pad nanoseconds."""
    expected = "000000013"
    actual = fake_filehandler.pad_nanoseconds(13)
    np.testing.assert_string_equal(actual, expected)


def test_nanoseconds_index():
    """Test nanosecond column being after seconds."""
    expected = UALF2_COLUMN_NAMES.index("nanosecond")
    actual = UALF2_COLUMN_NAMES.index("second") + 1
    np.testing.assert_array_equal(actual, expected)


def test_column_names_length():
    """Test correct number of column names."""
    expected = 30
    actual = len(UALF2_COLUMN_NAMES)
    np.testing.assert_equal(actual, expected)


@pytest.fixture
def fake_scene(fake_file):
    """Create fake file for tests."""
    from satpy import Scene

    scn = Scene(reader="gld360_ualf2", filenames=[fake_file])
    return scn


def test_scene_attributes(fake_scene):
    """Test for correct start and end times."""
    np.testing.assert_equal(fake_scene.start_time, TEST_START_TIME)
    np.testing.assert_equal(fake_scene.end_time, TEST_END_TIME)


def test_scene_load(fake_scene):
    """Test data loading through Scene-object."""
    fake_scene.load(["time", "latitude", "longitude"])
    assert "time" in fake_scene
    assert "latitude" in fake_scene
    assert "longitude" in fake_scene


def test_area_(fake_scene):
    """Test correct area instance type."""
    from pyresample.geometry import SwathDefinition

    fake_scene.load(["time"])
    assert isinstance(fake_scene["time"].attrs["area"], SwathDefinition)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Base modifier classes and utilities."""

from satpy.composites.core import CompositeBase


class ModifierBase(CompositeBase):
    """Base class for all modifiers.

    A modifier in Satpy is a class that takes one input DataArray to be
    changed along with zero or more other input DataArrays used to perform
    these changes. The result of a modifier typically has a lot of the same
    metadata (name, units, etc) as the original DataArray, but the data is
    different. A modified DataArray can be differentiated from the original
    DataArray by the `modifiers` property of its `DataID`.

    See the :class:`~satpy.composites.core.CompositeBase` class for information
    on the similar concept of "compositors".

    """

    def __call__(self, datasets, optional_datasets=None, **info):
        """Generate a modified copy of the first provided dataset."""
        raise NotImplementedError()



# Copyright (c) 2019-2023 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for the FCI L2 products in NetCDF4 format."""

import datetime as dt
import logging
from contextlib import suppress

import netCDF4
import numpy as np
import xarray as xr
from pyresample import geometry

from satpy._compat import cached_property
from satpy.area import get_area_def
from satpy.readers.core._geos_area import get_geos_area_naming, make_ext
from satpy.readers.core.eum import get_service_mode
from satpy.readers.core.fci import platform_name_translate
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger(__name__)

CHUNK_SIZE = get_legacy_chunk_size()

SSP_DEFAULT = 0.0


class FciL2CommonFunctions(object):
    """Shared operations for file handlers."""

    @property
    def spacecraft_name(self):
        """Return spacecraft name."""
        return platform_name_translate.get(
            self.nc.attrs["platform"], self.nc.attrs["platform"])

    @property
    def sensor_name(self):
        """Return instrument name."""
        return self.nc.attrs["data_source"].lower()

    @property
    def ssp_lon(self):
        """Return longitude at subsatellite point."""
        try:
            return float(self.nc["mtg_geos_projection"].attrs["longitude_of_projection_origin"])
        except (KeyError, AttributeError):
            logger.warning(f"ssp_lon could not be obtained from file content, using default value "
                           f"of {SSP_DEFAULT} degrees east instead")
            return SSP_DEFAULT

    def _get_global_attributes(self, product_type="pixel"):
        """Create a dictionary of global attributes to be added to all datasets.

        Returns:
            dict: A dictionary of global attributes.
                filename: name of the product file
                spacecraft_name: name of the spacecraft
                ssp_lon: longitude of subsatellite point
                sensor: name of sensor
                platform_name: name of the platform
            Only for AMVs product:
                channel: channel at which the AMVs have been retrieved
                time_parameters: dictionary of time attributes (currently only wind_time)

        """
        attributes = {
            "filename": self.filename,
            "spacecraft_name": platform_name_translate.get(self.spacecraft_name, self.spacecraft_name),
            "ssp_lon": self.ssp_lon,
            "sensor": self.sensor_name,
            "platform_name": platform_name_translate.get(self.spacecraft_name, self.spacecraft_name)
        }

        if product_type=="amv":
            attributes["channel"] = self.filename_info["channel"]
            attributes["time_parameters"] = {}
            attributes["time_parameters"]["wind_time"] = self.wind_time
        return attributes

    def _set_attributes(self, variable, dataset_info, product_type="pixel"):
        """Set dataset attributes."""
        if product_type in ["pixel", "segmented"]:
            if product_type == "pixel":
                xdim, ydim = "number_of_columns", "number_of_rows"
            elif product_type == "segmented":
                xdim, ydim = "number_of_FoR_cols", "number_of_FoR_rows"

            if dataset_info["nc_key"] not in ["product_quality",
                                              "product_completeness",
                                              "product_timeliness"]:
                variable = variable.swap_dims({ydim: "y", xdim: "x"})

        variable.attrs.setdefault("units", None)
        if "unit" in variable.attrs:
            # Need to convert this attribute to the expected satpy entry
            variable.attrs.update({"units": variable.attrs["unit"]})
            del variable.attrs["unit"]

        variable.attrs.update(dataset_info)
        variable.attrs.update(self._get_global_attributes(product_type=product_type))

        import_enum_information = dataset_info.get("import_enum_information", False)
        if import_enum_information:
            variable = self._add_flag_values_and_meanings(self.filename, dataset_info["nc_key"], variable)

        if variable.attrs["units"] == "none":
            variable.attrs.update({"units": None})

        return variable

    @staticmethod
    def _add_flag_values_and_meanings(filename, key, variable):
        """Build flag values and meaning from enum datatype."""
        nc_dataset = netCDF4.Dataset(filename, "r")
        # This currently assumes a flat netCDF file
        data_type = nc_dataset.variables[key].datatype
        if hasattr(data_type, "enum_dict"):
            enum = data_type.enum_dict
            flag_values = []
            flag_meanings = []
            for meaning, value in enum.items():
                flag_values.append(value)
                flag_meanings.append(meaning)

            variable.attrs["flag_values"] = flag_values
            variable.attrs["flag_meanings"] = flag_meanings
            nc_dataset.close()

        return variable

    def _slice_dataset(self, variable, dataset_info, dimensions):
        """Slice data if dimension layers have been provided in yaml-file."""
        slice_dict = {dim: dataset_info[dim_id] for (dim, dim_id) in dimensions.items()
                      if dim_id in dataset_info.keys() and dim in variable.dims}
        for dim, dim_ind in slice_dict.items():
            logger.debug(f"Extracting {dimensions[dim]}-index {dim_ind} from dimension {dim!r}.")
        variable = variable.sel(slice_dict)

        return variable

    @staticmethod
    def _mask_data(variable, fill_value):
        """Set fill_values, as defined in yaml-file, to NaN.

        Set data points in variable to NaN if they are equal to fill_value
        or any of the values in fill_value if fill_value is a list.
        """
        if not isinstance(fill_value, list):
            fill_value = [fill_value]

        for val in fill_value:
            variable = variable.where(variable != val).astype("float32")

        return variable

    def __del__(self):
        """Close the NetCDF file that may still be open."""
        with suppress(AttributeError, OSError):
            self.nc.close()


class FciL2NCFileHandler(FciL2CommonFunctions, BaseFileHandler):
    """Reader class for FCI L2 products in NetCDF4 format."""

    def __init__(self, filename, filename_info, filetype_info, with_area_definition=True):
        """Open the NetCDF file with xarray and prepare for dataset reading."""
        super().__init__(filename, filename_info, filetype_info)

        # Use xarray's default netcdf4 engine to open the fileq
        self.nc = xr.open_dataset(
            self.filename,
            decode_cf=True,
            mask_and_scale=True,
            chunks={
                "number_of_columns": CHUNK_SIZE,
                "number_of_rows": CHUNK_SIZE
            }
        )

        if with_area_definition is False:
            logger.info("Setting `with_area_defintion=False` has no effect on pixel-based products.")

        # Read metadata which are common to all datasets
        self.nlines = self.nc["y"].size
        self.ncols = self.nc["x"].size
        self._projection = self.nc["mtg_geos_projection"]
        self.multi_dims = {"maximum_number_of_layers": "layer", "number_of_vis_channels": "vis_channel_id"}

    def get_area_def(self, key):
        """Return the area definition."""
        try:
            return self._area_def
        except AttributeError:
            raise NotImplementedError

    def get_dataset(self, dataset_id, dataset_info):
        """Get dataset using the nc_key in dataset_info."""
        var_key = dataset_info["nc_key"]
        par_name = dataset_info["name"]
        logger.debug("Reading in file to get dataset with key %s.", var_key)

        try:
            variable = self.nc[var_key]
        except KeyError:
            logger.warning("Could not find key %s in NetCDF file, no valid Dataset created", var_key)
            return None

        # Compute the area definition
        if var_key not in ["product_quality", "product_completeness", "product_timeliness"]:
            self._area_def = self._compute_area_def(dataset_id)

        if any(dim_id in dataset_info.keys() for dim_id in self.multi_dims.values()):
            variable = self._slice_dataset(variable, dataset_info, self.multi_dims)

        if par_name == "retrieved_cloud_optical_thickness":
            variable = self.get_total_cot(variable)

        if dataset_info["file_type"] == "nc_fci_test_clm":
            variable = self._decode_clm_test_data(variable, dataset_info)

        if "fill_value" in dataset_info:
            variable = self._mask_data(variable, dataset_info["fill_value"])

        variable = self._set_attributes(variable, dataset_info)

        return variable

    @staticmethod
    def _decode_clm_test_data(variable, dataset_info):
        if dataset_info["nc_key"] != "cloud_mask_cmrt6_test_result":
            variable = variable.astype("uint32")
            variable.values = (variable.values >> dataset_info["extract_byte"] << 31 >> 31).astype("int8")

        return variable

    def _compute_area_def(self, dataset_id):
        """Compute the area definition.

        Returns:
            A pyresample AreaDefinition object containing the area definition.

        """
        area_extent = self._get_area_extent()
        area_naming, proj_dict = self._get_proj_area(dataset_id)
        area_def = geometry.AreaDefinition(
            area_naming["area_id"],
            area_naming["description"],
            "",
            proj_dict,
            self.ncols,
            self.nlines,
            area_extent)

        return area_def

    def _get_area_extent(self):
        """Calculate area extent of dataset."""
        # Load and convert x/y coordinates to degrees as required by the make_ext function
        x = self.nc["x"]
        y = self.nc["y"]
        x_deg = np.degrees(x)
        y_deg = np.degrees(y)

        # Select the extreme points and calcualte area extent (not: these refer to pixel center)
        ll_x, ur_x = -x_deg.values[0], -x_deg.values[-1]
        ll_y, ur_y = y_deg.values[-1], y_deg.values[0]
        h = float(self._projection.attrs["perspective_point_height"])
        area_extent_pixel_center = make_ext(ll_x, ur_x, ll_y, ur_y, h)

        # Shift area extent by half a pixel to get the area extent w.r.t. the dataset/pixel corners
        scale_factor = (x[1:] - x[0:-1]).values.mean()
        res = abs(scale_factor) * h
        area_extent = tuple(i + res / 2 if i > 0 else i - res / 2 for i in area_extent_pixel_center)

        return area_extent

    def _get_proj_area(self, dataset_id):
        """Extract projection and area information."""
        # Read the projection data from the mtg_geos_projection variable
        a = float(self._projection.attrs["semi_major_axis"])
        h = float(self._projection.attrs["perspective_point_height"])

        # Some L2PF test data files have a typo in the keyname for the inverse flattening parameter. Use a default value
        # as fallback until all L2PF test files are correctly formatted.
        rf = float(self._projection.attrs.get("inverse_flattening", 298.257223563))

        res = dataset_id["resolution"]

        area_naming_input_dict = {"platform_name": "mtg",
                                  "instrument_name": "fci",
                                  "resolution": res,
                                  }

        area_naming = get_geos_area_naming({**area_naming_input_dict,
                                            **get_service_mode("fci", self.ssp_lon)})

        proj_dict = {"a": a,
                     "lon_0": self.ssp_lon,
                     "h": h,
                     "rf": rf,
                     "proj": "geos",
                     "units": "m",
                     "sweep": "y"}

        return area_naming, proj_dict

    @staticmethod
    def get_total_cot(variable):
        """Sum the cloud optical thickness from the two OCA layers.

        The optical thickness has to be transformed to linear space before adding the values from the two layers. The
        combined/total optical thickness is then transformed back to logarithmic space.
        """
        attrs = variable.attrs
        variable = 10 ** variable
        variable = variable.fillna(0.)
        variable = variable.sum(dim="maximum_number_of_layers", keep_attrs=True)
        variable = variable.where(variable != 0., np.nan)
        variable = np.log10(variable)
        variable.attrs = attrs

        return variable


class FciL2NCSegmentFileHandler(FciL2CommonFunctions, BaseFileHandler):
    """Reader class for FCI L2 Segmented products in NetCDF4 format."""

    def __init__(self, filename, filename_info, filetype_info, with_area_definition=False):
        """Open the NetCDF file with xarray and prepare for dataset reading."""
        super().__init__(filename, filename_info, filetype_info)
        # Use xarray's default netcdf4 engine to open the file
        self.nc = xr.open_dataset(
            self.filename,
            decode_cf=True,
            mask_and_scale=True,
            chunks={
                "number_of_FoR_cols": CHUNK_SIZE,
                "number_of_FoR_rows": CHUNK_SIZE
            }
        )

        # Read metadata which are common to all datasets
        self.nlines = self.nc["number_of_FoR_rows"].size
        self.ncols = self.nc["number_of_FoR_cols"].size
        self.with_adef = with_area_definition
        self.multi_dims = {
            "number_of_categories": "category_id", "number_of_channels": "channel_id",
            "number_of_vis_channels": "vis_channel_id", "number_of_ir_channels": "ir_channel_id",
            "number_test": "test_id",
        }

    def get_area_def(self, key):
        """Return the area definition."""
        try:
            return self._area_def
        except AttributeError:
            raise NotImplementedError

    def get_dataset(self, dataset_id, dataset_info):
        """Get dataset using the nc_key in dataset_info."""
        var_key = dataset_info["nc_key"]
        logger.debug("Reading in file to get dataset with key %s.", var_key)

        try:
            variable = self.nc[var_key]
        except KeyError:
            logger.warning("Could not find key %s in NetCDF file, no valid Dataset created", var_key)
            return None

        if any(dim_id in dataset_info.keys() for dim_id in self.multi_dims.values()):
            variable = self._slice_dataset(variable, dataset_info, self.multi_dims)

        if self.with_adef and var_key not in ["longitude", "latitude",
                                              "product_quality", "product_completeness", "product_timeliness"]:
            self._area_def = self._construct_area_def(dataset_id)

            # coordinates are not relevant when returning data with an AreaDefinition
            if "coordinates" in dataset_info.keys():
                del dataset_info["coordinates"]

        if "fill_value" in dataset_info:
            variable = self._mask_data(variable, dataset_info["fill_value"])

        variable = self._set_attributes(variable, dataset_info, product_type="segmented")

        return variable

    def _construct_area_def(self, dataset_id):
        """Construct the area definition.

        Returns:
            A pyresample AreaDefinition object containing the area definition.

        """
        res = dataset_id["resolution"]

        area_naming_input_dict = {"platform_name": "mtg",
                                  "instrument_name": "fci",
                                  "resolution": res,
                                  }

        area_naming = get_geos_area_naming({**area_naming_input_dict,
                                            **get_service_mode("fci", self.ssp_lon)})

        # Construct area definition from standardized area definition.
        stand_area_def = get_area_def(area_naming["area_id"])

        if (stand_area_def.width != self.ncols) | (stand_area_def.height != self.nlines):
            raise NotImplementedError("Unrecognised AreaDefinition.")

        mod_area_extent = self._modify_area_extent(stand_area_def.area_extent)

        area_def = geometry.AreaDefinition(
            stand_area_def.area_id,
            stand_area_def.description,
            "",
            stand_area_def.crs,
            stand_area_def.width,
            stand_area_def.height,
            mod_area_extent)

        return area_def

    @staticmethod
    def _modify_area_extent(stand_area_extent):
        """Modify area extent to macth satellite projection.

        Area extent has to be modified since the L2 products are stored with the south-east
        in the upper-right corner (as opposed to north-east in the standardized area definitions).
        """
        ll_x, ll_y, ur_x, ur_y = stand_area_extent
        ll_y *= -1.
        ur_y *= -1.
        area_extent = tuple([ll_x, ll_y, ur_x, ur_y])

        return area_extent


class FciL2NCAMVFileHandler(FciL2CommonFunctions, BaseFileHandler):
    """Reader class for FCI L2 AMV products in NetCDF4 format."""

    def __init__(self, filename, filename_info, filetype_info):
        """Open the NetCDF file with xarray and prepare for dataset reading."""
        super().__init__(filename, filename_info, filetype_info)

    @cached_property
    def nc(self):
        """Read the file."""
        return xr.open_dataset(
            self.filename,
            decode_cf=True,
            mask_and_scale=True,
            chunks={
                "number_of_images": CHUNK_SIZE,
                "number_of_winds": CHUNK_SIZE
            }
        )

    @property
    def wind_time(self):
        """Return wind time as a datetime object.

        This quantity represents the start_time of the second image used
        for the AMV tracking, and is already present as a wind_time dataset
        in units of Seconds since 2000-01-01 00:00:00.0.
        """
        ref = dt.datetime(2000,1,1,0,0,0)
        try:
            secs_since_ref = self.nc["wind_time"].data.item()
            return ref + dt.timedelta(seconds=secs_since_ref)
        except KeyError:
            logger.warning("wind_time dataset not found. Setting wind_time to None.")
            return None

    def get_dataset(self, dataset_id, dataset_info):
        """Get dataset using the nc_key in dataset_info."""
        var_key = dataset_info["nc_key"]
        logger.debug("Reading in file to get dataset with key %s.", var_key)

        try:
            variable = self.nc[var_key]
        except KeyError:
            logger.warning("Could not find key %s in NetCDF file, no valid Dataset created", var_key)
            return None

        # Manage the attributes of the dataset
        variable = self._set_attributes(variable, dataset_info, product_type="amv")

        return variable



# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Compositors using lookup tables."""

from __future__ import annotations

import logging

import numpy as np
import xarray as xr

from .core import CompositeBase, GenericCompositor

LOG = logging.getLogger(__name__)


class CategoricalDataCompositor(CompositeBase):
    """Compositor used to recategorize categorical data using a look-up-table.

    Each value in the data array will be recategorized to a new category defined in
    the look-up-table using the original value as an index for that look-up-table.

    Example:
        data = [[1, 3, 2], [4, 2, 0]]
        lut = [10, 20, 30, 40, 50]
        res = [[20, 40, 30], [50, 30, 10]]
    """

    def __init__(self, name, lut=None, **kwargs):  # noqa: D417
        """Get look-up-table used to recategorize data.

        Args:
            lut (list): a list of new categories. The lenght must be greater than the
                        maximum value in the data array that should be recategorized.
        """
        self.lut = np.array(lut)
        super(CategoricalDataCompositor, self).__init__(name, **kwargs)

    def _update_attrs(self, new_attrs):
        """Modify name and add LUT."""
        new_attrs["name"] = self.attrs["name"]
        new_attrs["composite_lut"] = list(self.lut)

    @staticmethod
    def _getitem(block, lut):
        return lut[block]

    def __call__(self, projectables, **kwargs):
        """Recategorize the data."""
        if len(projectables) != 1:
            raise ValueError("Can't have more than one dataset for a categorical data composite")

        data = projectables[0].astype(int)
        res = data.data.map_blocks(
            self._getitem,
            self.lut,
            dtype=self.lut.dtype,
            meta=np.ndarray((), dtype=self.lut.dtype),
        )

        new_attrs = data.attrs.copy()
        self._update_attrs(new_attrs)

        return xr.DataArray(res, dims=data.dims, attrs=new_attrs, coords=data.coords)


class ColormapCompositor(GenericCompositor):
    """A compositor that uses colormaps.

    .. warning::

        Deprecated since Satpy 0.39.

    This compositor is deprecated.  To apply a colormap, use a
    :class:`satpy.composites.core.SingleBandCompositor` composite with a
    :func:`~satpy.enhancements.colormap.colorize` or
    :func:`~satpy.enhancements.colormap.palettize` enhancement instead.
    For example, to make a ``cloud_top_height`` composite based on a dataset
    ``ctth_alti`` palettized by ``ctth_alti_pal``, the composite would be::

      cloud_top_height:
        compositor: !!python/name:satpy.composites.core.SingleBandCompositor
        prerequisites:
        - ctth_alti
        tandard_name: cloud_top_height

    and the enhancement::

      cloud_top_height:
        standard_name: cloud_top_height
        operations:
        - name: palettize
          method: !!python/name:satpy.enhancements.colormap.palettize
          kwargs:
            palettes:
              - dataset: ctth_alti_pal
                color_scale: 255
                min_value: 0
                max_value: 255
    """

    @staticmethod
    def build_colormap(palette, dtype, info):
        """Create the colormap from the `raw_palette` and the valid_range.

        Colormaps come in different forms, but they are all supposed to have
        color values between 0 and 255. The following cases are considered:

        - Palettes comprised of only a list of colors. If *dtype* is uint8,
          the values of the colormap are the enumeration of the colors.
          Otherwise, the colormap values will be spread evenly from the min
          to the max of the valid_range provided in `info`.
        - Palettes that have a palette_meanings attribute. The palette meanings
          will be used as values of the colormap.

        """
        from trollimage.colormap import Colormap

        squeezed_palette = np.asanyarray(palette).squeeze() / 255.0
        cmap = Colormap.from_array_with_metadata(
                palette,
                dtype,
                color_scale=255,
                valid_range=info.get("valid_range"),
                scale_factor=info.get("scale_factor", 1),
                add_offset=info.get("add_offset", 0))

        return cmap, squeezed_palette

    def __call__(self, projectables, **info):
        """Generate the composite."""
        if len(projectables) != 2:
            raise ValueError("Expected 2 datasets, got %d" %
                             (len(projectables), ))
        data, palette = projectables

        colormap, palette = self.build_colormap(palette, data.dtype, data.attrs)

        channels = self._apply_colormap(colormap, data, palette)
        return self._create_composite_from_channels(channels, data)

    def _create_composite_from_channels(self, channels, template):
        mask = self._get_mask_from_data(template)
        channels = [self._create_masked_dataarray_like(channel, template, mask) for channel in channels]
        res = super(ColormapCompositor, self).__call__(channels, **template.attrs)
        res.attrs["_FillValue"] = np.nan
        return res

    @staticmethod
    def _get_mask_from_data(data):
        fill_value = data.attrs.get("_FillValue", np.nan)
        if np.isnan(fill_value):
            mask = data.notnull()
        else:
            mask = data != data.attrs["_FillValue"]
        return mask

    @staticmethod
    def _create_masked_dataarray_like(array, template, mask):
        return xr.DataArray(array.reshape(template.shape),
                            dims=template.dims, coords=template.coords,
                            attrs=template.attrs).where(mask)


class ColorizeCompositor(ColormapCompositor):
    """A compositor colorizing the data, interpolating the palette colors when needed.

    .. warning::

        Deprecated since Satpy 0.39.  See the :class:`ColormapCompositor`
        docstring for documentation on the alternative.
    """

    @staticmethod
    def _apply_colormap(colormap, data, palette):
        del palette
        return colormap.colorize(data.data.squeeze())


class PaletteCompositor(ColormapCompositor):
    """A compositor colorizing the data, not interpolating the palette colors.

    .. warning::

        Deprecated since Satpy 0.39.  See the :class:`ColormapCompositor`
        docstring for documentation on the alternative.
    """

    @staticmethod
    def _apply_colormap(colormap, data, palette):
        channels, colors = colormap.palettize(data.data.squeeze())
        channels = channels.map_blocks(_insert_palette_colors, palette, dtype=palette.dtype,
                                       meta=np.ndarray((), dtype=palette.dtype),
                                       new_axis=2, chunks=list(channels.chunks) + [palette.shape[1]])
        return [channels[:, :, i] for i in range(channels.shape[2])]


def _insert_palette_colors(channels, palette):
    channels = palette[channels]
    return channels
